package net.sourceforge.rapidprops;

import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.lang.reflect.Method;
import java.util.HashMap;
import java.util.Map;
import java.util.Properties;

import net.sourceforge.rapidprops.Props.Handler;

public class PropsBuilder {
    private static final Method defineClassMethod;
    static {
        Method m;
        try {
            m = ClassLoader.class.getDeclaredMethod("defineClass", new Class[] {String.class, byte[].class, Integer.TYPE, Integer.TYPE});
            m.setAccessible(true);
        } catch (NoSuchMethodException nsme) {
            m = null;
        }
        defineClassMethod = m;
    }

	@SuppressWarnings("unchecked")
	public static <T> T createFromMap(Class<T> propsInterface, Map<?,?> map) {
		try {
			boolean trimKeys = true;
			boolean trimValues = true;
			Handler keyHandler = new Props.NoHandling();
			Handler valueHandler = new Props.NoHandling();

			Props propsAnnotation = propsInterface.getAnnotation(Props.class);
			if(propsAnnotation != null) {
				trimKeys = propsAnnotation.trimKeys();
				trimValues = propsAnnotation.trimValues();
				keyHandler = propsAnnotation.keyHandler().getConstructor().newInstance();
				valueHandler = propsAnnotation.valueHandler().getConstructor().newInstance();
			}
			Map<String, String> handledMap = new HashMap<String, String>();
			for(Map.Entry<?,?> entry : map.entrySet()) {
				Object origKey = entry.getKey();
				String key = null;
				if(origKey != null) {
					key = String.valueOf(origKey);
					if(trimKeys) key = key.trim();
				}
				key = keyHandler.handle(key);

				Object origValue = entry.getValue();
				String value = null;
				if(origValue != null) {
					value = String.valueOf(origValue);
					if(trimValues) value = value.trim();
				}
				value = valueHandler.handle(value);

				handledMap.put(key, value);
			}
			PropsHelper<T> helper = new PropsHelper<T>(propsInterface, handledMap);
			byte[] byteCode = helper.createByteCode();
			String clsName = helper.getClassNameDot();
			ClassLoader cl = Thread.currentThread().getContextClassLoader();
			Class<T> implClass = (Class<T>)defineClassMethod.invoke(cl, new Object[]{clsName, byteCode, 0, byteCode.length});
			T implementation = implClass.newInstance();
			return implementation;
		} catch(Throwable e) {
			throw new RapidPropsException("Failed to implement interface " + propsInterface.getName(), e);
		}
	}

	public static <T> T createFromFile(Class<T> propsInterface, String filePath) throws IOException {
		Properties props = new Properties();
		props.load(new FileInputStream(filePath));
		return createFromMap(propsInterface, props);
	}

	public static <T> T createFromResource(Class<T> propsInterface, String resourceName) throws IOException {
		Properties props = new Properties();
		InputStream inputStream = propsInterface.getResourceAsStream(resourceName);
		props.load(inputStream);
		return createFromMap(propsInterface, props);
	}
}
